//    Copyright (C) 2010  Luis Alfonso Arce González, ajaest[@]gmail.com
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

package net.ajaest.jdk.core.main;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.swing.SwingUtilities;

import net.ajaest.jdk.core.auxi.LastVersionChecker;
import net.ajaest.jdk.core.exporters.Exporter;
import net.ajaest.jdk.core.exporters.KanjiPlainTextExporter;
import net.ajaest.jdk.core.main.ExceptionHandler.KNOWN_EXCEPTIONS;
import net.ajaest.jdk.data.auxi.SimilarKanjiStrokeDemuxer;
import net.ajaest.jdk.data.dict.KanjiDict;
import net.ajaest.jdk.data.dict.auxi.KanjiDatabaseInfo;
import net.ajaest.jdk.data.dict.query.KanjiExpression;
import net.ajaest.jdk.data.dict.query.QueryAdaptor;
import net.ajaest.jdk.data.dict.sort.KanjiOrdering;
import net.ajaest.jdk.data.kanji.KanjiTag;
import net.ajaest.lib.utils.SystemEnums.ISO639ー1;

//TODO: JavaDoc 
public class JavaDiKt implements Runnable {

	// Program engine
	private LastVersionChecker lvc;
	private JDKOptions jdkOpt;
	private JDKGUIEngine jdkGui;
	private KanjiDict kd;
	private List<Exporter<KanjiTag>> exporters;
	private SimilarKanjiStrokeDemuxer sksdmx;

	// Program fields
	public static final String VERSION = "1.1.4 beta";
	public static final String PROJECT = "Mirai";

	private String lastVersion = null;



	public JavaDiKt(JDKOptions jdkOpt) {

		try {
			lvc = new LastVersionChecker(new URL("http://ajaest.net/javadikt/lastVersion.txt"));
			lvc.checkVersion();
		} catch (MalformedURLException e1) {
		}

		this.jdkOpt = jdkOpt;

		try {
			kd = new KanjiDict(jdkOpt.getKanjiFile(), jdkOpt.getTreeFile());

		} catch (FileNotFoundException e) {
			ExceptionHandler.handleException(e, KNOWN_EXCEPTIONS.INVALID_KANJI_OR_STROKE_FILE_PATH);
		} catch (IOException e) {
			// TODO:
		}

		exporters = new ArrayList<Exporter<KanjiTag>>();
		exporters.add(new KanjiPlainTextExporter(jdkOpt.getLanguage().toLowerCase()));

		jdkGui = new JDKGUIEngine(this);

	}

	@Override
	public void run() {
		SwingUtilities.invokeLater(jdkGui);
	}

	public List<KanjiTag> executeQuery(KanjiExpression currentKE) {

		List<KanjiTag> results;
		
		if(currentKE==null){
			results = new ArrayList<KanjiTag>();
			Messages.localePrintln("**JavaDiKt: WARINING - Null query");
		}
		else
			results = kd.executeQuery(adaptQuery(currentKE));

		Messages.localePrintln("**JavaDiKt: Query Executed: " + currentKE);
		return results;
	}

	@SuppressWarnings("static-access")
	private KanjiExpression adaptQuery(KanjiExpression ke) {

		QueryAdaptor qa = QueryAdaptor.groupPairs().demuxReadings().demuxGraphs(sksdmx, 5).meaningsInLanguaje(ISO639ー1.valueOf(jdkOpt.getLanguage().toUpperCase()));
		
		if (!jdkOpt.getExtendedIKanjiList()) {
			qa = qa.limitToJôyô();
		}

		return qa.adapt(ke);
	}

	public List<Integer> executeRefQuery(KanjiExpression ke) {

		ke = adaptQuery(ke);

		List<Integer> result = kd.executeRefQuery(ke);
		System.out.println("**JavaDiKt: Reference Query Executed: " + ke);
		kd.sortKanjiRefs(result,new KanjiOrdering().sort_by_frequency().from_least_to_greatest());
		
		return result;
	}

	public JDKOptions getOptions() {
		return jdkOpt;
	}

	public void saveOptions(String lang, String kanjiDictpath, String strokeDictPath, Boolean extInfo, Boolean extKanji, Boolean romaji) {

		jdkOpt.setLanguage(lang);
		jdkOpt.setKanjiFile(new File(kanjiDictpath));
		jdkOpt.setTreeFile(new File(strokeDictPath));
		jdkOpt.setExtendedInfo(extInfo);
		jdkOpt.setFindInExtended(extKanji);
		jdkOpt.setRomanized(romaji);
		jdkOpt.saveCurrentConfig();
	}

	public void setExporters(List<Exporter<KanjiTag>> exporters) {

		for (Exporter<KanjiTag> exp : exporters) {
			exp.setLanguaje(jdkOpt.getLanguage());
		}

		this.exporters = exporters;
	}

	public List<Exporter<KanjiTag>> getExporters() {
		return exporters;
	}

	public JDKGUIEngine getJDKGuiEngine() {
		return jdkGui;
	}
	public KanjiDatabaseInfo getKanjiDatabaseInfo() {
		return kd.getDatabaseInfo();
	}

	public SimilarKanjiStrokeDemuxer getSimilarGraphDemuxer() {
		return sksdmx;
	}

	public void setSimilarStrokeDemuxer(SimilarKanjiStrokeDemuxer sksdmx) {
		this.sksdmx = sksdmx;
	}

	public boolean isLastVersion() {

		if (lastVersion == null)
			lastVersion = getLastVersion();

		return lastVersion.equals(VERSION);
	}

	public String getLastVersion() {

		if (lastVersion == null) {
			if (lvc.isChecking()) {
				// order the version checker to write the last version info in
				// an file. The file version will be read the next time the
				// program is launched
				lvc.setLastVersionFile(new File("config/ver.txt"));
				lvc.writeToFile(true);
				// reads version info from a previous session
				lastVersion = getVersionFromFile("config/ver.txt");
			} else {
				if (lvc.readSuccessfull())
					lastVersion = lvc.getVersion();
				else {
					// reads version info from a previous session
					lastVersion = getVersionFromFile("config/ver.txt");
				}
			}
		}

		return lastVersion;
	}

	private String getVersionFromFile(String path) {

		String version;

		try {
			BufferedReader br = new BufferedReader(new FileReader(path));
			version = br.readLine();
			br.close();
		} catch (Exception e) {
			version = VERSION;
		}

		return version;
	}

	public KanjiTag getKanjiByRef(Integer kanjiRef) {

		System.out.println("**JavaDiKt: Retrieving kanji by unicode value: " + kanjiRef);
		return kd.getKanjiByUnicode(kanjiRef);
	}
}
