//    Copyright (C) 2010  Luis Alfonso Arce González, ajaest[@]gmail.com
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

package net.ajaest.jdk.data.dict.auxi;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.NavigableSet;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.zip.ZipFile;

import net.ajaest.jdk.core.main.ExceptionHandler;
import net.ajaest.jdk.core.main.ExceptionHandler.KNOWN_EXCEPTIONS;
import net.ajaest.jdk.data.auxi.AllowedStrokeLineEnum;
import net.ajaest.jdk.data.dict.query.IndexPairQAbout;
import net.ajaest.jdk.data.dict.query.IntegerValueQAbout;
import net.ajaest.jdk.data.dict.query.StringPairQAbout;
import net.ajaest.jdk.data.dict.query.StringValueQAbout;
import net.ajaest.jdk.data.dict.query.ValueQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.KanjiGraphQAbout;
import net.ajaest.lib.data.SequenceTree;
import net.ajaest.lib.string.Strings;

//TODO: JavaDoc
public class TreeContainer {


	// trees
	private TreeSet<Integer> 									kanjiUnicodeTree;
	
	private SequenceTree<AllowedStrokeLineEnum, Integer> 		kanjiGraphTree;
	private SequenceTree<Integer, Integer> 						kanjiJisCodeTree; // 1
	private SequenceTree<Integer, Integer> 						kanjiSKIPCodeTree;
	private SequenceTree<Character, Integer> 					kanjiDeRooCodeTree;
	private SequenceTree<Character, Integer> 					kanjiFourCornerTree;
	private SequenceTree<Character, Integer> 					kanjiMeaningTree; // 2
	private SequenceTree<Character, Integer> 					kanjiReadingTree; // 3
	private SequenceTree<Character, Integer> 					kanjiSpahnHadamitzkyCodeTree;
	private SequenceTree<Character, Integer> 					kanjiVariantTree; // 4
	
	private HashMap<String, Set<Integer>> 						kanjiJisCharset; // 1
	private HashMap<String, Set<Integer>> 						kanjiDicNameTree; // 5
	private HashMap<String, Set<Integer>> 						kanjiMeaningLangTree; // 2
	private HashMap<String, Set<Integer>> 						kanjiReadingTypeTree; // 3
	private HashMap<String, Set<Integer>> 						kanjiVariantTypeTree; // 4
	
	private TreeMap<Integer, Integer> 							kanjiFrequencyTreeSet;
	private TreeMap<Integer, Set<Integer>> 						kanjiClassicRadicalTree;
	private TreeMap<Integer, Set<Integer>> 						kanjiGradeTree;
	private TreeMap<Integer, Set<Integer>> 						kanjiJLPTLevelTree;
	private TreeMap<Integer, Set<Integer>> 						kanjiNelsonRadicalTree;
	private TreeMap<Integer, Set<Integer>> 						kanjiStrokeCountTree;
	private TreeMap<Integer, Set<Integer>> 						kanjiStrokeMiscountsTree;
	private TreeMap<Integer, HashMap<String,Set<Integer>>>		kanjiDicReferenceTree;
	private TreeMap<Integer, Set<Integer>> 						kanjiDicReferencePlainTree;

	// #############################################################
	// ############# Constructor methods ###########################
	// #############################################################

	public TreeContainer(File zobj) {
		loadTrees(zobj);
	}

	// #############################################################
	// ############# Public methods ################################
	// #############################################################

	public Set<Integer> getRefs(ValueQAbout<?> vqa) {

		Set<Integer> result = new HashSet<Integer>();
		
		if (vqa instanceof IntegerValueQAbout) {

			IntegerValueQAbout ivaq = (IntegerValueQAbout) vqa;
			switch (ivaq.getFieldEnum()) {
				case KANJI_CLASSIC_RADICAL :
					result.addAll(getFromTreeMap(kanjiClassicRadicalTree, ivaq.getQueryCase(), ivaq.getValue()));
					break;
				case KANJI_CLASSIC_NELSON :
					result.addAll(getFromTreeMap(kanjiNelsonRadicalTree, ivaq.getQueryCase(), ivaq.getValue()));
					break;
				case KANJI_STROKE_MISCOUNT :
					result.addAll(getFromTreeMap(kanjiStrokeMiscountsTree, ivaq.getQueryCase(), ivaq.getValue()));
					break;
				case KANJI_DIC_INDEX : // TODO: remove one of this elements
				case KANJI_DIC_REFERENCE :
					result.addAll(getFromTreeMap(kanjiDicReferencePlainTree, ivaq.getQueryCase(), ivaq.getValue()));
					break;
				case KANJI_FREQUENCY :
					result.addAll(getFromTreeMap(kanjiFrequencyTreeSet, ivaq.getQueryCase(), ivaq.getValue()));
					break;
				case KANJI_GRADE :
					result.addAll(getFromTreeMap(kanjiGradeTree, ivaq.getQueryCase(), ivaq.getValue()));
					break;
				case KANJI_JLPT_LEVEL :
					result.addAll(getFromTreeMap(kanjiJLPTLevelTree, ivaq.getQueryCase(), ivaq.getValue()));
					break;
				case KANJI_STROKE_COUNT :
					result.addAll(getFromTreeMap(kanjiStrokeCountTree, ivaq.getQueryCase(), ivaq.getValue()));
					break;
				case KANJI_UNICODE_VALUE :
					result.addAll(getFromTreeSet(kanjiUnicodeTree, ivaq.getQueryCase(), ivaq.getValue()));
					break;
				default :
					throw new IllegalArgumentException("Not allowed field:" + vqa.getFieldEnum());
			}
		} else if (vqa instanceof StringValueQAbout) {

			List<Integer> seqi;
			LinkedList<Character> seqc;
			StringValueQAbout svqa = (StringValueQAbout) vqa;
			switch (svqa.getFieldEnum()) {
				case KANJI_JIS_CHARSET :
					result.addAll(getFromHashMap(kanjiJisCharset, svqa.getQueryCase(), svqa.getValue()));
					break;
				case KANJI_DIC_NAME :
					result.addAll(getFromHashMap(kanjiDicNameTree, svqa.getQueryCase(), svqa.getValue()));
					break;
				case KANJI_MEANING_LANGUAGE :
					result.addAll(getFromHashMap(kanjiMeaningLangTree, svqa.getQueryCase(), svqa.getValue()));
					break;
				case KANJI_READING_TYPE :
					result.addAll(getFromHashMap(kanjiReadingTypeTree, svqa.getQueryCase(), svqa.getValue()));
					break;
				case KANJI_VARIANT_TYPE :
					result.addAll(getFromHashMap(kanjiVariantTypeTree, svqa.getQueryCase(), svqa.getValue()));
					break;
				case KANJI_JIS_CODE :
					// TODO: JIS to integer in KanjiTag.java
					seqi = new ArrayList<Integer>();
					for (String c : svqa.getValue().split("-"))
						seqi.add(new Integer(c));
					if (seqi.size() == 2) {
						seqi.add(seqi.get(1));
						seqi.set(1, seqi.get(0));
						seqi.set(0, 1);
					}
					result.addAll(getFromSequenceTree(kanjiJisCodeTree, svqa.getQueryCase(), seqi));
					break;
				case KANJI_SKIP :
					seqi = new ArrayList<Integer>();
					for (String c : svqa.getValue().split("-"))
						seqi.add(new Integer(c));
					result.addAll(getFromSequenceTree(kanjiSKIPCodeTree, svqa.getQueryCase(), seqi));
					break;
				case KANJI_DE_ROO :
					seqc = Strings.toCharacters(svqa.getValue());
					result.addAll(getFromSequenceTree(kanjiDeRooCodeTree, svqa.getQueryCase(), seqc));
					break;
				case KANJI_FOUR_CORNER :
					seqc = Strings.toCharacters(svqa.getValue());
					result.addAll(getFromSequenceTree(kanjiFourCornerTree, svqa.getQueryCase(), seqc));
					break;
				case KANJI_SPAHN_HADAMITZKY :
					seqc = Strings.toCharacters(svqa.getValue());
					result.addAll(getFromSequenceTree(kanjiSpahnHadamitzkyCodeTree, svqa.getQueryCase(), seqc));
					break;
				case KANJI_MEANING :
					seqc = Strings.toCharacters(svqa.getValue());
					result.addAll(getFromSequenceTree(kanjiMeaningTree, svqa.getQueryCase(), seqc));
					break;
				case KANJI_READING :
					seqc = Strings.toCharacters(svqa.getValue());
					result.addAll(getFromSequenceTree(kanjiReadingTree, svqa.getQueryCase(), seqc));
					break;
				case KANJI_VARIANT_INDEX :
					seqc = Strings.toCharacters(svqa.getValue());
					result.addAll(getFromSequenceTree(kanjiVariantTree, svqa.getQueryCase(), seqc));
					break;
				default :
					throw new IllegalArgumentException("Not allowed field:" + vqa.getFieldEnum());

			}
		} else if (vqa instanceof IndexPairQAbout) {
			IndexPairQAbout ipqa = (IndexPairQAbout) vqa;
			switch (ipqa.getFieldEnum()) {
				case PAIR_DIC :
					result.addAll(getFromDicTree(kanjiDicReferenceTree, ipqa.getQueryCase(), ipqa));
					break;
				default :
					throw new IllegalArgumentException("Not allowed field:" + vqa.getFieldEnum());
			}
		} else if (vqa instanceof StringPairQAbout) {
			List<Integer> seqi;
			LinkedList<Character> seqc;

			StringPairQAbout spqa = (StringPairQAbout) vqa;
			switch (spqa.getFieldEnum()) {
				case PAIR_JIS :
					// TODO: JIS to integer in KanjiTag.java
					seqi = new ArrayList<Integer>();
					for (String c : spqa.getValue().getSecond().split("-"))
						seqi.add(new Integer(c));
					if (seqi.size() == 2) {
						seqi.add(seqi.get(1));
						seqi.set(1, seqi.get(0));
						seqi.set(0, 1);
					}
					result.addAll(getFromSequenceTree(kanjiJisCodeTree, spqa.getQueryCase(), seqi));
					result.retainAll(getFromHashMap(kanjiJisCharset, spqa.getQueryCase(), spqa.getValue().getFirst()));
					break;
				case PAIR_MEANING :
					seqi = new ArrayList<Integer>();
					seqc = Strings.toCharacters(spqa.getValue().getSecond());
					seqc.add('#');
					seqc.addAll(Strings.toCharacters(spqa.getValue().getFirst()));
					result.addAll(getFromSequenceTree(kanjiMeaningTree, spqa.getQueryCase(), seqc));
					break;
				case PAIR_READING :
					seqi = new ArrayList<Integer>();
					seqc = Strings.toCharacters(spqa.getValue().getSecond());
					seqc.add('#');
					seqc.addAll(Strings.toCharacters(spqa.getValue().getFirst()));
					result.addAll(getFromSequenceTree(kanjiReadingTree, spqa.getQueryCase(), seqc));
					break;
				case PAIR_VARIANT :
					seqi = new ArrayList<Integer>();
					seqc = Strings.toCharacters(spqa.getValue().getSecond());
					seqc.add('#');
					seqc.addAll(Strings.toCharacters(spqa.getValue().getFirst()));
					result.addAll(getFromSequenceTree(kanjiVariantTree, spqa.getQueryCase(), seqc));
					break;
				default :
					throw new IllegalArgumentException("Not allowed field:" + vqa.getFieldEnum());
			}

		} else if (vqa instanceof KanjiGraphQAbout) {
			KanjiGraphQAbout kgqa = (KanjiGraphQAbout) vqa;
			switch (vqa.getFieldEnum()) {
				case KANJI_GRAPH :
					result.addAll(getFromSequenceTree(kanjiGraphTree, kgqa.getQueryCase(), kgqa.getValue().getLineSequence()));
					break;
				default :
					throw new IllegalArgumentException("Not allowed field:" + vqa.getFieldEnum());
			}
		} else {
			throw new IllegalArgumentException("Not allowed field:" + vqa.getFieldEnum());
		}
		return result;
	}

	public SequenceTree<Character, Integer> getKanjiDeRooCodeTree() {
		return kanjiDeRooCodeTree;
	}

	public SequenceTree<Character, Integer> getKanjiFourCornerTree() {
		return kanjiFourCornerTree;
	}

	public HashMap<String, Set<Integer>> getKanjiDicNameTree() {
		return kanjiDicNameTree;
	}

	public TreeMap<Integer, Integer> getKanjiFrequencyTreeSet() {
		return kanjiFrequencyTreeSet;
	}

	public TreeMap<Integer, Set<Integer>> getKanjiClassicRadicalTree() {
		return kanjiClassicRadicalTree;
	}

	public TreeMap<Integer, HashMap<String, Set<Integer>>> getKanjiDicReferenceTree() {
		return kanjiDicReferenceTree;
	}

	public TreeMap<Integer, Set<Integer>> getKanjiDicReferencePlainTree() {
		return kanjiDicReferencePlainTree;
	}

	public TreeSet<Integer> getKanjiUnicodeTree() {
		return kanjiUnicodeTree;
	}

	public SequenceTree<AllowedStrokeLineEnum, Integer> getKanjiGraphTree() {
		return kanjiGraphTree;
	}

	public SequenceTree<Integer, Integer> getKanjiJisCodeTree() {
		return kanjiJisCodeTree;
	}

	public SequenceTree<Integer, Integer> getKanjiSKIPCodeTree() {
		return kanjiSKIPCodeTree;
	}

	public SequenceTree<Character, Integer> getKanjiMeaningTree() {
		return kanjiMeaningTree;
	}

	public SequenceTree<Character, Integer> getKanjiReadingTree() {
		return kanjiReadingTree;
	}

	public SequenceTree<Character, Integer> getKanjiSpahnHadamitzkyCodeTree() {
		return kanjiSpahnHadamitzkyCodeTree;
	}

	public SequenceTree<Character, Integer> getKanjiVariantTree() {
		return kanjiVariantTree;
	}

	public HashMap<String, Set<Integer>> getKanjiJisCharset() {
		return kanjiJisCharset;
	}

	public HashMap<String, Set<Integer>> getKanjiMeaningLangTree() {
		return kanjiMeaningLangTree;
	}

	public HashMap<String, Set<Integer>> getKanjiReadingTypeTree() {
		return kanjiReadingTypeTree;
	}

	public HashMap<String, Set<Integer>> getKanjiVariantTypeTree() {
		return kanjiVariantTypeTree;
	}

	public TreeMap<Integer, Set<Integer>> getKanjiGradeTree() {
		return kanjiGradeTree;
	}

	public TreeMap<Integer, Set<Integer>> getKanjiJLPTLevelTree() {
		return kanjiJLPTLevelTree;
	}

	public TreeMap<Integer, Set<Integer>> getKanjiNelsonRadicalTree() {
		return kanjiNelsonRadicalTree;
	}

	public TreeMap<Integer, Set<Integer>> getKanjiStrokeCountTree() {
		return kanjiStrokeCountTree;
	}

	public TreeMap<Integer, Set<Integer>> getKanjiStrokeMiscountsTree() {
		return kanjiStrokeMiscountsTree;
	}

	// #############################################################
	// ############# Private methods ###############################
	// #############################################################

	private Collection<? extends Integer> getFromDicTree(TreeMap<Integer, HashMap<String, Set<Integer>>> map, QueryCaseEnum qc, IndexPairQAbout qa) {

		Set<Integer> dicResult;
		Set<Integer> result = new HashSet<Integer>();
		Collection<HashMap<String, Set<Integer>>> plainedDics;
		switch (qc) {
			case EQUALS :
				result.addAll(map.get(qa.getValue().getSecond()).get(qa.getValue().getFirst()));
				break;
			case NOT_EQUALS :
				for (Set<Integer> refs : kanjiDicReferencePlainTree.values())
					result.addAll(refs);
				result.removeAll(map.get(qa.getValue().getSecond()).get(qa.getValue().getFirst()));
				break;
			case GREATER_OR_EQUALS_THAN :
				plainedDics = map.tailMap(qa.getValue().getSecond(), true).values();
				for (HashMap<String, Set<Integer>> dicRefs : plainedDics) {
					dicResult = dicRefs.get(qa.getValue().getFirst());
					if (dicResult != null)
						result.addAll(dicResult);
				}
				break;
			case GREATER_THAN :
				plainedDics = map.tailMap(qa.getValue().getSecond(), false).values();
				for (HashMap<String, Set<Integer>> dicRefs : plainedDics) {
					dicResult = dicRefs.get(qa.getValue().getFirst());
					if (dicResult != null)
						result.addAll(dicResult);
				}
				break;
			case LESS_OR_EQUALS_THAN :
				plainedDics = map.headMap(qa.getValue().getSecond(), true).values();
				for (HashMap<String, Set<Integer>> dicRefs : plainedDics) {
					dicResult = dicRefs.get(qa.getValue().getFirst());
					if (dicResult != null)
						result.addAll(dicResult);
				}
				break;
			case LESS_THAN :
				plainedDics = map.headMap(qa.getValue().getSecond(), false).values();
				for (HashMap<String, Set<Integer>> dicRefs : plainedDics) {
					dicResult = dicRefs.get(qa.getValue().getFirst());
					if (dicResult != null)
						result.addAll(dicResult);
				}
				break;
			case NULL :
				result.addAll(kanjiUnicodeTree);// TODO:slow
				for (Set<Integer> s : kanjiDicReferencePlainTree.values())
					result.removeAll(s);
				break;
			default :
				throw new IllegalArgumentException("Not allowed query case:" + qc);
		}

		return result;
	}

	private Collection<Integer> getFromTreeSet(NavigableSet<Integer> set, QueryCaseEnum qc, Integer value) {

		Set<Integer> result = new HashSet<Integer>();

		switch (qc) {
			case EQUALS :
				if (set.contains(value))
					result.add(value);
				break;
			case NOT_EQUALS :
				result.addAll(kanjiUnicodeTree);
				result.remove(value);
				break;
			case GREATER_OR_EQUALS_THAN :
				result.addAll(set.tailSet(value, true));
				break;
			case GREATER_THAN :
				result.addAll(set.tailSet(value, false));
				break;
			case LESS_OR_EQUALS_THAN :
				result.addAll(set.headSet(value, true));
				break;
			case LESS_THAN :
				result.addAll(set.headSet(value, false));
				break;
			case NULL :
				result.addAll(kanjiUnicodeTree);
				result.removeAll(set);
				break;
			default :
				throw new IllegalArgumentException("Not allowed query case:" + qc);
		}

		return result;
	}
	
	private <E> Collection<Integer> getFromTreeMap(TreeMap<Integer, Integer> map, QueryCaseEnum qc, Integer value) {

		Set<Integer> result = new HashSet<Integer>();

		switch (qc) {
			case EQUALS :
				result.add(map.get(value));
				break;
			case NOT_EQUALS :
				result.addAll(kanjiUnicodeTree);
				result.remove(map.get(value));
				break;
			case GREATER_OR_EQUALS_THAN :
				result.addAll(map.tailMap(value, true).values());
				break;
			case GREATER_THAN :
				result.addAll(map.tailMap(value, false).values());
				break;
			case LESS_OR_EQUALS_THAN :
				result.addAll(map.headMap(value, true).values());
				break;
			case LESS_THAN :
				result.addAll(map.headMap(value, true).values());
				break;
			case NULL :
				result.addAll(kanjiUnicodeTree);
				result.removeAll(map.values());
				break;
			default :
				throw new IllegalArgumentException("Not allowed query case:" + qc);
		}

		return result;
	}


	private <E> Collection<Integer> getFromTreeMap(TreeMap<E, Set<Integer>> map, QueryCaseEnum qc, E value) {

		Set<Integer> result = new HashSet<Integer>();

		switch (qc) {
			case EQUALS :
				result.addAll(map.get(value));
				break;
			case NOT_EQUALS :
				result.addAll(kanjiUnicodeTree);
				result.removeAll(map.get(value));
				break;
			case GREATER_OR_EQUALS_THAN :
				for (Set<Integer> val : map.tailMap(value, true).values())
					result.addAll(val);

				break;
			case GREATER_THAN :
				for (Set<Integer> val : map.tailMap(value, false).values())
					result.addAll(val);

				break;
			case LESS_OR_EQUALS_THAN :
				for (Set<Integer> val : map.headMap(value, true).values())
					result.addAll(val);

				break;
			case LESS_THAN :
				for (Set<Integer> val : map.headMap(value, true).values())
					result.addAll(val);
				break;
			case NULL :
				result.addAll(kanjiUnicodeTree);
				result.removeAll(map.values());
				break;
			default :
				throw new IllegalArgumentException("Not allowed query case:" + qc);
		}

		return result;
	}

	private <E> Collection<Integer> getFromSequenceTree(SequenceTree<E, Integer> tree, QueryCaseEnum qc, List<E> seq) {

		Set<Integer> result = new HashSet<Integer>();
		
		switch (qc) {
			case EQUALS :
				result.addAll(tree.search(seq));
				break;
			case NOT_EQUALS :
				result.addAll(kanjiUnicodeTree);
				result.removeAll(tree.search(seq));
				break;
			case NULL :
				result.addAll(kanjiUnicodeTree);
				for (SequenceTree<E, Integer> subTree : tree)
					result.removeAll(subTree.getFinalNodes());
				break;
			default :
				throw new IllegalArgumentException("Not allowed query case:" + qc);
		}

		return result;
	}

	private <E> Collection<Integer> getFromHashMap(HashMap<E, Set<Integer>> map, QueryCaseEnum qc, E value) {

		Set<Integer> result = new HashSet<Integer>();

		switch (qc) {
			case EQUALS :
				result.addAll(map.get(value));
				break;
			case NOT_EQUALS :
				result.addAll(kanjiUnicodeTree);
				result.removeAll(map.get(value));
				break;
			case NULL :
				result.addAll(kanjiUnicodeTree);
				result.removeAll(map.values());
			default :
				throw new IllegalArgumentException("Not allowed query case:" + qc);
		}

		return result;
	}

	@SuppressWarnings("unchecked")
	private void loadTrees(File zobj) {

		ZipFile zip;
		try {
			zip = new ZipFile(zobj);

			kanjiSKIPCodeTree = (SequenceTree<Integer, Integer>) readObjectFromZip(zip, "kanjiSKIPCodeTree");

			kanjiFourCornerTree = (SequenceTree<Character, Integer>) readObjectFromZip(zip, "kanjiFourCornerTree");

			kanjiSpahnHadamitzkyCodeTree = (SequenceTree<Character, Integer>) readObjectFromZip(zip, "kanjiSpahnHadamitzkyCodeTree");

			kanjiDeRooCodeTree = (SequenceTree<Character, Integer>) readObjectFromZip(zip, "kanjiDeRooCodeTree");

			kanjiGraphTree = (SequenceTree<AllowedStrokeLineEnum, Integer>) readObjectFromZip(zip, "kanjiGraphTree");

			kanjiJisCharset = (HashMap<String, Set<Integer>>) readObjectFromZip(zip, "kanjiJisCharset");

			kanjiMeaningTree = (SequenceTree<Character, Integer>) readObjectFromZip(zip, "kanjiMeaningTree");

			kanjiReadingTree = (SequenceTree<Character, Integer>) readObjectFromZip(zip, "kanjiReadingTree");

			kanjiJisCodeTree = (SequenceTree<Integer, Integer>) readObjectFromZip(zip, "kanjiJisCodeTree");

			kanjiVariantTree = (SequenceTree<Character, Integer>) readObjectFromZip(zip, "kanjiVariantTree");

			kanjiUnicodeTree = (TreeSet<Integer>) readObjectFromZip(zip, "kanjiUnicodeTree");

			kanjiClassicRadicalTree = (TreeMap<Integer, Set<Integer>>) readObjectFromZip(zip, "kanjiClassicRadicalTree");

			kanjiDicNameTree = (HashMap<String, Set<Integer>>) readObjectFromZip(zip, "kanjiDicNameTree");

			kanjiDicReferenceTree = (TreeMap<Integer, HashMap<String,Set<Integer>>>) readObjectFromZip(zip, "kanjiDicReferenceTree");

			kanjiDicReferencePlainTree = new TreeMap<Integer, Set<Integer>>(); //TODO: remove, slow and unnecesary 
			Set<Integer> plainedRefs;
			for (Integer key : kanjiDicReferenceTree.keySet()) {
				for (Set<Integer> dicRefs : kanjiDicReferenceTree.get(key).values()) {
					plainedRefs = kanjiDicReferencePlainTree.get(key);
					if (plainedRefs == null) {
						plainedRefs = new HashSet<Integer>();
						kanjiDicReferencePlainTree.put(key, plainedRefs);
					}
					plainedRefs.addAll(dicRefs);
				}
			}

			kanjiFrequencyTreeSet = (TreeMap<Integer, Integer>) readObjectFromZip(zip, "kanjiFrequencyTreeSet");

			kanjiGradeTree = (TreeMap<Integer, Set<Integer>>) readObjectFromZip(zip, "kanjiGradeTree");

			kanjiJLPTLevelTree = (TreeMap<Integer, Set<Integer>>) readObjectFromZip(zip, "kanjiJLPTLevelTree");

			kanjiMeaningLangTree = (HashMap<String, Set<Integer>>) readObjectFromZip(zip, "kanjiMeaningLangTree");

			kanjiNelsonRadicalTree = (TreeMap<Integer, Set<Integer>>) readObjectFromZip(zip, "kanjiNelsonRadicalTree");

			kanjiReadingTypeTree = (HashMap<String, Set<Integer>>) readObjectFromZip(zip, "kanjiReadingTypeTree");

			kanjiStrokeCountTree = (TreeMap<Integer, Set<Integer>>) readObjectFromZip(zip, "kanjiStrokeCountTree");

			kanjiStrokeMiscountsTree = (TreeMap<Integer, Set<Integer>>) readObjectFromZip(zip, "kanjiStrokeMiscountsTree");

			kanjiVariantTypeTree = (HashMap<String, Set<Integer>>) readObjectFromZip(zip, "kanjiVariantTypeTree");

		} catch (IOException e) {
			ExceptionHandler.handleException(e, KNOWN_EXCEPTIONS.TREE_IOEXCEPTION);
		} catch (ClassNotFoundException e) {
			ExceptionHandler.handleException(e, KNOWN_EXCEPTIONS.CORRUPTED_TREE);
		}
	}

	/**
	 * Returns the first object serialized into a file of the zip file
	 */
	private Object readObjectFromZip(ZipFile zip, String name) throws IOException, ClassNotFoundException {

		Object obj = null;

		InputStream zis;
		ObjectInputStream ois;
		try {
			zis = zip.getInputStream(zip.getEntry(name));
			ois = new ObjectInputStream(new BufferedInputStream(zis));

			obj = ois.readObject();
			zis.close();
			ois.close();

		} catch (RuntimeException e) {
			System.err.println("Failed when reading \"" + name + "\" zip file");
			throw e;
		} catch (IOException e) {
			System.err.println("Failed when reading \"" + name + "\" zip file");
			throw e;
		}

		return obj;
	}


}

