//    Copyright (C) 2010  Luis Alfonso Arce González, ajaest[@]gmail.com
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

package net.ajaest.lib.data;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

//TODO: update diagram
/**
 * Search tree by sequence. This class allows to sort a list of node objects
 * using a unique sequence of other objects. Every level of the tree represents
 * all the node objects that have in common a particular list in specific order
 * of other objects. This class can use either {@code Arraylist} or {@code
 * LinkedList} implementations for final nodes list. It's recommended to use
 * {@code Arraylist} implementation for fixed trees and {@code LinkedList}
 * implementation for dynamic trees.
 * 
 * @author Luis Alfonso Arce González
 * 
 * @param <S>
 *            the class of the sequence objects
 * @param <E>
 *            the class of the stored nodes objects
 */
public class OrderedSequenceTree<S, E> {

	private OrderedSequenceTree<S, E> parent;
	private S sequenceValue;
	private Map<S, OrderedSequenceTree<S, E>> subtrees;
	private List<E> finalNodes;
	private boolean isLinked;

	/**
	 * Creates a ordered sequence tree using either {@code Arraylist} or {@code
	 * LinkedList} implementation for final nodes list. It's recommended to use
	 * {@code Arraylist} implementations for fixed trees and {@code LinkedList}
	 * implementation for dynamic trees.
	 * 
	 * @param linked
	 *            if it's true, object will use {@code LinkedList}
	 *            implementation. Object will use {@code ArrayList}
	 *            implementation if it's false.
	 */
	public OrderedSequenceTree(boolean linked) {

		isLinked = linked;

		subtrees = new HashMap<S, OrderedSequenceTree<S, E>>();
		if (linked)
			finalNodes = new LinkedList<E>();
		else
			finalNodes = new ArrayList<E>();
		sequenceValue = null;
		parent = null;

	}

	/**
	 * Creates needed sequence tree if necessary in order to add a final node to
	 * the tree. If the {@code sequence.size()==0}, the final node is added to
	 * this tree.
	 * 
	 * @param sequence
	 *            the sorting sequence
	 * @param finalNode
	 *            the final node
	 */
	public void add(List<S> sequence, E finalNode) {

		if (sequence.size() > 0) {

			OrderedSequenceTree<S, E> st = subtrees.get(sequence.get(0));

			if (st == null) {
				st = new OrderedSequenceTree<S, E>(isLinked);
				st.parent = this;
				st.sequenceValue = sequence.get(0);
				subtrees.put(sequence.get(0), st);
			}

			if (sequence.size() == 1) {
				st.addFinalNode(finalNode);
			} else {
				st.add(sequence.subList(1, sequence.size()), finalNode);
			}
		} else {
			this.addFinalNode(finalNode);
		}

	}

	/**
	 * Returns an unmodifiable map that maps subtrees from this tree by his
	 * sequence value.
	 * 
	 * @return a map with this tree's subtrees
	 */
	public Map<S, OrderedSequenceTree<S, E>> getSubTrees() {

		return Collections.unmodifiableMap(subtrees);
	}

	/**
	 * 
	 * @return the list of final nodes related to this tree
	 */
	public List<E> getFinalNodes() {

		return finalNodes;
	}

	/**
	 * Returns the subtree whose sequence value equals the specified one.
	 * 
	 * @param sequenceValue
	 *            the sequence value of the requested subtree
	 * @return the requested subtree, null otherwise
	 */
	public OrderedSequenceTree<S, E> getSubTree(S sequenceValue) {

		return subtrees.get(sequenceValue);
	}

	/**
	 * 
	 * @return the sequence value that maps this tree
	 */
	public S getSequenceValue() {

		return sequenceValue;
	}

	/**
	 * 
	 * @return the tree parent of this tree
	 */
	public OrderedSequenceTree<S, E> getParent() {

		return parent;
	}

	/**
	 * Adds a subtree to this tree using either provided {@code sequenceValue}
	 * or {@code subtree.sequenceValue}. If {@code sequenceValue} don't equals
	 * {@code null}, this value will be used to map the subtree. {@code
	 * subtree.sequenceValue} will be used to map the subtree if {@code
	 * sequenceValue} equals {@code null} only if {@code subtree.sequenceValue}
	 * does not equals {@code null}. If both sequence values equals {@code null}
	 * exception will be thrown. <br>
	 * Keep in mind that if subtree was nested in another tree before adding to
	 * this, previous parent information and sequence value will be overwritten
	 * in order to nest it to this structure.
	 * 
	 * @param sequenceValue
	 *            order mapping value, if null {@code sequenceValue} will be
	 *            used
	 * @param subtree
	 *            subtree to be added
	 * @throws NullPointerException
	 *             if {@code sequenceValue} and {@code subtree.sequenceValue}
	 *             are both null
	 */
	public void addSubTree(S sequenceValue, OrderedSequenceTree<S, E> subtree) {

		S tempSV = sequenceValue != null ? sequenceValue : subtree.sequenceValue;

		if (tempSV == null)
			throw new NullPointerException("sortvalue and subtree.getSortValue() equals null, no sort information");

		if (subtrees.get(tempSV) != null)
			subtrees.get(tempSV).subtrees.get(tempSV).finalNodes.addAll(subtree.getFinalNodes());
		else {
			subtree.parent = this;
			subtree.sequenceValue = tempSV;
			subtrees.put(tempSV, subtree);
		}
	}

	/**
	 * Replaces the element at the specified position in this list with the
	 * specified element.
	 * 
	 * @param index
	 *            index of the element to replace
	 * @param element
	 *            element to be stored at the specified position
	 * @return the element previously at the specified position
	 * @throws IndexOutOfBoundsException
	 */
	public E setFinalNode(int index, E element) {
		return finalNodes.set(index, element);
	}

	/**
	 * Appends the specified element to the end of this list.
	 * 
	 * @param node
	 *            element to be appended to this list
	 * @return <tt>true</tt> (as specified by {@link Collection#add})
	 */
	public boolean addFinalNode(E node) {

		return finalNodes.add(node);
	}

	/**
	 * Removes the mapping for the specified key from this map if present.
	 * 
	 * @param sequenceValue
	 *            key whose mapping is to be removed from the map
	 * @return the previous value associated with <tt>key</tt>, or <tt>null</tt>
	 *         if there was no mapping for <tt>key</tt>.
	 */
	public OrderedSequenceTree<S, E> removeSubTree(S sequenceValue) {

		return subtrees.remove(sequenceValue);
	}

	/**
	 * Removes the element at the specified position in this list. Shifts any
	 * subsequent elements to the left (subtracts one from their indices).
	 * 
	 * @param index
	 *            the index of the element to be removed
	 * @return the element that was removed from the list
	 * @throws IndexOutOfBoundsException
	 */
	public E removeFinalNode(int index) {

		return finalNodes.remove(index);
	}

	/**
	 * Removes the first occurrence of the specified element from this list, if
	 * it is present. If the list does not contain the element, it is unchanged.
	 * More formally, removes the element with the lowest index <tt>i</tt> such
	 * that
	 * <tt>(o==null&nbsp;?&nbsp;get(i)==null&nbsp;:&nbsp;o.equals(get(i)))</tt>
	 * (if such an element exists). Returns <tt>true</tt> if this list contained
	 * the specified element (or equivalently, if this list changed as a result
	 * of the call).
	 * 
	 * @param node
	 *            element to be removed from this list, if present
	 * @return <tt>true</tt> if this list contained the specified element
	 */
	public boolean removeFinalNode(E node) {

		return finalNodes.remove(node);
	}

	/**
	 * Does a recursive search along subtrees of this tree in order to get final
	 * nodes that matches the specified sequence. If not matching final node is
	 * found, it returns an empty list.
	 * 
	 * @param sequence
	 *            sequence to be searched
	 * @return list of final nodes that matches the sequence
	 */
	public List<E> search(List<S> sequence) {

		if (sequence.size() == 0)
			return finalNodes;
		else {
			OrderedSequenceTree<S, E> temp = subtrees.get(sequence.get(0));
			if (temp != null) {
				if (sequence.size() == 1)
					return temp.finalNodes;
				else
					return temp.search(sequence.subList(1, sequence.size()));
			}
		}

		return new ArrayList<E>();

	}

	/**
	 * Says whether exists or not any final node in this tree referenced by the
	 * specified sequence
	 * 
	 * @param sequence
	 *            sequence to be checked
	 * @return true if {@code search(sequence)} does not returns an empty list,
	 *         false otherwise
	 */
	public boolean exits(List<S> sequence) {

		return search(sequence).size() != 0;

	}

	@Override
	public int hashCode() {
		// I wan't it to be super fast. I don't want recursive calculation of
		// subtrees hashcode, that's why I do not use parent and subtrees
		// TODO: 2010-12-12
		final int prime = 31;
		int result = 1;
		result = prime * result + ((finalNodes == null) ? 0 : finalNodes.hashCode());
		result = prime * result + (isLinked ? 1231 : 1237);
		result = prime * result + ((sequenceValue == null) ? 0 : sequenceValue.hashCode());
		return result;
	}

	/**
	 * To sequence tree equals if they have the same sequence value and final
	 * nodes.
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		OrderedSequenceTree<?, ?> other = (OrderedSequenceTree<?, ?>) obj;
		if (finalNodes == null) {
			if (other.finalNodes != null)
				return false;
		} else if (!finalNodes.equals(other.finalNodes))
			return false;
		if (isLinked != other.isLinked)
			return false;
		if (sequenceValue == null) {
			if (other.sequenceValue != null)
				return false;
		} else if (!sequenceValue.equals(other.sequenceValue))
			return false;
		return true;
	}
}
