//    Copyright (C) 2010  Luis Alfonso Arce González, ajaest[@]gmail.com
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

package net.ajaest.lib.data;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;

/**
 * Search tree by sequence. This class allows to sort a list of node objects
 * using a unique sequence an enumeration elements. Compared to
 * {@link SequenceTree}, this class is likely to be faster provided it
 * uses {@code EnumMap} implementation to look for child nodes.<br>
 * <br>
 * Every level of the tree represents all the node objects that can be retrieved
 * through the same enum sequence. This class can use either {@code ArrayList}
 * or {@code LinkedList} implementations for final nodes list. It's recommended
 * to use {@code ArrayList} implementation for fixed trees and {@code
 * LinkedList} implementation for dynamic trees.<br>
 * <br>
 * You must be aware that the node circularity of the tree isn't checked when a
 * new node is added. Circular connection between nodes could cause unexpected
 * behavior.<br>
 * <br>
 * Keep in mind too that the serialization process of an instance of this class
 * won't retrieve information about the parent structure. This means that the
 * {@code readObject()} method will always assign {@code null} to the parent
 * field of the tree's root node.
 * 
 * @author Luis Alfonso Arce González
 * 
 * @param <S>
 *            the class of the sequence objects
 * @param <E>
 *            the class of the stored nodes objects
 */
public class SequenceEnumTree<S extends Enum<S>, E> extends
		SequenceTree<S, E> implements Externalizable {

	private static final long serialVersionUID = 8886707074769021876L;

	private Class<S> enumClass;

	public SequenceEnumTree() {
	}
	
	public SequenceEnumTree(boolean linked, Class<S> keyType) {
		this.enumClass = keyType;
		super.init(linked);
	}
	

	@Override
	protected Map<S, SequenceTree<S, E>> createSubTreeMapInstance() {

		return new EnumMap<S, SequenceTree<S, E>>(enumClass);
	}
	
	@SuppressWarnings({ "unchecked" })
	@Override
	protected SequenceTree<S, E> readTreeSequentially(ObjectInput in)
			throws IOException, ClassNotFoundException {
		
		Class<S> enumC = (Class<S>) in.readObject(); 
		SequenceEnumTree<S, E> oset = new SequenceEnumTree<S, E>(in.readBoolean(),enumC);
		oset.serialID = in.readInt();
		oset.sequenceValue = (S) in.readObject();
		oset.finalNodes = (List<E>) in.readObject();
		
		return oset;
	}
	
	@Override
	protected void writeTreeSequentially(ObjectOutput out,
			SequenceTree<S, E> ost) throws IOException {

		out.writeObject(enumClass);
		out.writeBoolean(ost.isLinked);
		out.writeInt(ost.serialID);
		out.writeObject(ost.sequenceValue);
		out.writeObject(ost.finalNodes);
	}

	@Override
	protected void readAndAssignToThis(ObjectInput in) throws IOException,
			ClassNotFoundException {
		SequenceEnumTree<S, E> readNode = (SequenceEnumTree<S, E>) readTreeSequentially(in);
		this.parent = null;
		this.enumClass = readNode.enumClass;
		this.finalNodes = readNode.finalNodes;
		this.isLinked = readNode.isLinked;
		this.sequenceValue = readNode.sequenceValue;
		this.serialID = readNode.serialID;
	}

}
