//    Copyright (C) 2010  Luis Alfonso Arce González, ajaest[@]gmail.com
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

package net.ajaest.lib.data;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Queue;
import java.util.Set;

/**
 * Search tree by sequence. This class allows to sort a list of node objects
 * using a unique sequence of other objects. Every level of the tree represents
 * all the node objects that can be retrieved through the same object sequence.
 * This class can use either {@code ArrayList} or {@code LinkedList}
 * implementations for final nodes list. It's recommended to use {@code
 * ArrayList} implementation for fixed trees and {@code LinkedList}
 * implementation for dynamic trees.<br>
 * <br>
 * You must be aware that the node circularity of the tree isn't checked when a
 * new node is added. Circular connection between nodes could cause unexpected
 * behavior.<br>
 * <br>
 * Keep in mind too that if the serialization process of an instance of this
 * class won't retrieve information about the parent structure. This means that
 * the {@code readObject()} method will always assign {@code null} to the parent
 * field of the tree's root node.
 * 
 * @author Luis Alfonso Arce González
 * 
 * @param <S>
 *            the class of the sequence objects
 * @param <E>
 *            the class of the stored nodes objects
 */
public class SequenceTree<S, E> implements Externalizable,
		Iterable<SequenceTree<S, E>> {

	private static final long 	serialVersionUID = 3402834272156144704L;
	protected static Integer 	serialIDseed = 0;
	/**
	 * Used to keep node reference in serialization
	 */
	protected int				serialID;
	
	protected 					SequenceTree<S, E> parent;
	protected 					S sequenceValue;
	/**
	 * Serialization of big trees causes stack overflow because of recursive
	 * reflection, so maps are serialized horizontally instead of deeply.
	 * 
	 * @see #writeObject
	 * @see #readObject
	 */
	transient protected			Map<S, SequenceTree<S, E>> subtrees;
	protected List<E> 			finalNodes;
	protected boolean 			isLinked;

	/**
	 * Creates a sequence tree using either the{@code LinkedList} implementation
	 * for final nodes list.
	 */
	public SequenceTree(){
	}

	/**
	 * Creates a sequence tree using either {@code ArrayList} or {@code
	 * LinkedList} implementation for final nodes list. It's recommended to use
	 * {@code ArrayList} implementations for fixed trees and {@code LinkedList}
	 * implementation for dynamic trees.
	 * 
	 * @param linked
	 *            if it's true, object will use {@code LinkedList}
	 *            implementation. Object will use {@code ArrayList}
	 *            implementation if it's false.
	 */
	public SequenceTree(boolean linked) {
		init(linked);
	}

	/**
	 * Me Method extracted in order to allow post-construction initialization in
	 * inherited classes(For example, in {@link SequenceEnumTree}
	 * 
	 * @param linked
	 */
	protected void init(boolean linked) {
		isLinked = linked;

		serialID = serialIDseed;
		serialIDseed++;

		subtrees = createSubTreeMapInstance();

		if (linked)
			finalNodes = new LinkedList<E>();
		else
			finalNodes = new ArrayList<E>();

		sequenceValue = null;
		parent = null;
	}
	
	/**
	 * Creates needed sequence tree if necessary in order to add a final node to
	 * the tree. If the {@code sequence.size()==0}, the final node is added to
	 * this tree.
	 * 
	 * @param sequence
	 *            the sorting sequence
	 * @param finalNode
	 *            the final node
	 */
	public void add(List<S> sequence, E finalNode) {

		if (sequence.size() > 0) {

			SequenceTree<S, E> parent = this;
			SequenceTree<S, E> child = null;

			for (S seqElm : sequence) {
				child = parent.subtrees.get(seqElm);
				if (child == null) {
					child = new SequenceTree<S, E>(isLinked);
					child.parent = parent;
					child.sequenceValue = seqElm;
					parent.subtrees.put(seqElm, child);
				}
				parent = child;
			}

			child.addFinalNode(finalNode);

		} else {
			this.addFinalNode(finalNode);
		}

		// Recursive algorithm produces stack overflow easily
		// if (sequence.size() > 0) {
		//
		// OrderedSequenceTree<S, E> st = subtrees.get(sequence.get(0));
		//
		// if (st == null) {
		// st = new OrderedSequenceTree<S, E>(isLinked);
		// st.parent = this;
		// st.sequenceValue = sequence.get(0);
		// subtrees.put(sequence.get(0), st);
		// }
		//
		// if (sequence.size() == 1) {
		// st.addFinalNode(finalNode);
		// } else {
		// st.add(sequence.subList(1, sequence.size()), finalNode);
		// }
		// } else {
		// this.addFinalNode(finalNode);
		// }

	}

	/**
	 * Creates needed sequence tree if necessary in order to add a final node to
	 * the tree. If the {@code sequence.length==0}, the final node is added to
	 * this tree.
	 * 
	 * @param sequence
	 *            the sorting sequence
	 * @param finalNode
	 *            the final node
	 */
	public void add(S[] sequence, E finalNode) {

		if (sequence.length > 0) {

			SequenceTree<S, E> parent = this;
			SequenceTree<S, E> child = null;

			for (S seqElm : sequence) {
				child = parent.subtrees.get(seqElm);
				if (child == null) {
					child = new SequenceTree<S, E>(isLinked);
					child.parent = parent;
					child.sequenceValue = seqElm;
					parent.subtrees.put(seqElm, child);
				}
				parent = child;
			}

			child.addFinalNode(finalNode);

		} else {
			this.addFinalNode(finalNode);
		}
	}
	/**
	 * Returns an unmodifiable map that maps subtrees from this tree by his
	 * sequence value.
	 * 
	 * @return a map with this tree's subtrees
	 */
	public Map<S, SequenceTree<S, E>> getSubTrees() {

		return Collections.unmodifiableMap(subtrees);
	}

	/**
	 * 
	 * @return the list of final nodes related to this tree
	 */
	public List<E> getFinalNodes() {

		return finalNodes;
	}

	/**
	 * Returns the subtree whose sequence value equals the specified one.
	 * 
	 * @param sequenceValue
	 *            the sequence value of the requested subtree
	 * @return the requested subtree, null otherwise
	 */
	public SequenceTree<S, E> getSubTree(S sequenceValue) {

		return subtrees.get(sequenceValue);
	}

	/**
	 * 
	 * @return the sequence value that maps this tree
	 */
	public S getSequenceValue() {

		return sequenceValue;
	}

	/**
	 * 
	 * @return the tree parent of this tree
	 */
	public SequenceTree<S, E> getParent() {

		return parent;
	}

	/**
	 * Adds a subtree to this tree using either provided {@code sequenceValue}
	 * or {@code subtree.sequenceValue}. If {@code sequenceValue} don't equals
	 * {@code null}, this value will be used to map the subtree. {@code
	 * subtree.sequenceValue} will be used to map the subtree if {@code
	 * sequenceValue} equals {@code null} only if {@code subtree.sequenceValue}
	 * does not equals {@code null}. If both sequence values equals {@code null}
	 * exception will be thrown. <br>
	 * Keep in mind that if subtree was nested in another tree before adding to
	 * this, previous parent information and sequence value will be overwritten
	 * in order to nest it to this structure.
	 * 
	 * @param sequenceValue
	 *            order mapping value, if null {@code sequenceValue} will be
	 *            used
	 * @param subtree
	 *            subtree to be added
	 * @throws NullPointerException
	 *             if {@code sequenceValue} and {@code subtree.sequenceValue}
	 *             are both null
	 */
	public void addSubTree(S sequenceValue, SequenceTree<S, E> subtree) {

		S tempSV = sequenceValue != null ? sequenceValue : subtree.sequenceValue;

		if (tempSV == null)
			throw new NullPointerException("sortvalue and subtree.getSortValue() equals null, no sort information");

		if (subtrees.get(tempSV) != null)
			subtrees.get(tempSV).subtrees.get(tempSV).finalNodes.addAll(subtree.getFinalNodes());
		else {
			subtree.parent = this;
			subtree.sequenceValue = tempSV;
			subtrees.put(tempSV, subtree);
		}
	}

	/**
	 * Replaces the element at the specified position in this list with the
	 * specified element.
	 * 
	 * @param index
	 *            index of the element to replace
	 * @param element
	 *            element to be stored at the specified position
	 * @return the element previously at the specified position
	 * @throws IndexOutOfBoundsException
	 */
	public E setFinalNode(int index, E element) {
		return finalNodes.set(index, element);
	}

	/**
	 * Appends the specified element to the end of this list.
	 * 
	 * @param node
	 *            element to be appended to this list
	 * @return <tt>true</tt> (as specified by {@link Collection#add})
	 */
	public boolean addFinalNode(E node) {

		return finalNodes.add(node);
	}

	/**
	 * Removes the mapping for the specified key from this map if present.
	 * 
	 * @param sequenceValue
	 *            key whose mapping is to be removed from the map
	 * @return the previous value associated with <tt>key</tt>, or <tt>null</tt>
	 *         if there was no mapping for <tt>key</tt>.
	 */
	public SequenceTree<S, E> removeSubTree(S sequenceValue) {

		return subtrees.remove(sequenceValue);
	}

	/**
	 * Removes the element at the specified position in this list. Shifts any
	 * subsequent elements to the left (subtracts one from their indices).
	 * 
	 * @param index
	 *            the index of the element to be removed
	 * @return the element that was removed from the list
	 * @throws IndexOutOfBoundsException
	 */
	public E removeFinalNode(int index) {

		return finalNodes.remove(index);
	}

	/**
	 * Removes the first occurrence of the specified element from this list, if
	 * it is present. If the list does not contain the element, it is unchanged.
	 * More formally, removes the element with the lowest index <tt>i</tt> such
	 * that
	 * <tt>(o==null&nbsp;?&nbsp;get(i)==null&nbsp;:&nbsp;o.equals(get(i)))</tt>
	 * (if such an element exists). Returns <tt>true</tt> if this list contained
	 * the specified element (or equivalently, if this list changed as a result
	 * of the call).
	 * 
	 * @param node
	 *            element to be removed from this list, if present
	 * @return <tt>true</tt> if this list contained the specified element
	 */
	public boolean removeFinalNode(E node) {

		return finalNodes.remove(node);
	}

	/**
	 * Does a recursive search along subtrees of this tree in order to get final
	 * nodes that matches the specified sequence. If not matching final node is
	 * found, it returns an empty list.
	 * 
	 * @param sequence
	 *            sequence to be searched
	 * @return list of final nodes that matches the sequence
	 */
	public List<E> search(List<S> sequence) {

		if (sequence.size() == 0)
			return finalNodes;
		else {
			SequenceTree<S, E> temp = subtrees.get(sequence.get(0));
			if (temp != null) {
				if (sequence.size() == 1)
					return temp.finalNodes;
				else
					return temp.search(sequence.subList(1, sequence.size()));
			}
		}

		return new ArrayList<E>();

	}

	/**
	 * Does a recursive search along subtrees of this tree in order to get the
	 * first final node added to this tree that matches the specified sequence.
	 * If not matching final node is found, it returns null
	 * 
	 * @param sequence
	 *            sequence to be searched
	 * @return the final node that matches the sequence, null otherwise
	 */
	public E searchFirst(List<S> sequence) {

		E node;
		List<E> retrieved = search(sequence);

		if (retrieved.size() == 0)
			node = null;
		else
			node = retrieved.get(0);

		return node;
	}

	/**
	 * Does a recursive search along subtrees of this tree in order to get the
	 * first final node added to this tree that matches the specified sequence.
	 * If not matching final node is found, it returns null
	 * 
	 * @param sequence
	 *            sequence to be searched
	 * @return the final node that matches the sequence, null otherwise
	 */
	public E searchFirst(S[] sequence) {

		ArrayList<S> al = new ArrayList<S>();
		Collections.addAll(al, sequence);

		return searchFirst(al);
	}

	/**
	 * Does a recursive search along subtrees of this tree in order to get final
	 * nodes that matches the specified sequence. If not matching final node is
	 * found, it returns an empty list.
	 * 
	 * @param sequence
	 *            sequence to be searched
	 * @return list of final nodes that matches the sequence
	 */
	public List<E> search(S[] sequence) {

		ArrayList<S> al = new ArrayList<S>();
		Collections.addAll(al, sequence);

		return search(al);
	}


	/**
	 * Says whether exists or not any final node in this tree referenced by the
	 * specified sequence
	 * 
	 * @param sequence
	 *            sequence to be checked
	 * @return true if {@code search(sequence)} does not returns an empty list,
	 *         false otherwise
	 */
	public boolean exist(List<S> sequence) {

		return search(sequence).size() != 0;

	}

	/**
	 * Returns an iterator that iterates through all the nodes contained in this
	 * tree. No special ordering is used. The root isn't returned by {@code
	 * next()}.
	 */
	@Override
	public Iterator<SequenceTree<S, E>> iterator() {

		return new TreeIterator(this);
	}

	@Override
	public int hashCode() {
		// I wan't it to be super fast. I don't want recursive calculation of
		// subtrees hashcode, that's why I do not use parent and subtrees
		// TODO: 2010-12-12
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((finalNodes == null) ? 0 : finalNodes.hashCode());
		result = prime * result + (isLinked ? 1231 : 1237);
		result = prime * result
				+ ((sequenceValue == null) ? 0 : sequenceValue.hashCode());
		return result;
	}

	/**
	 * To sequence tree equals if they have the same sequence value and final
	 * nodes.
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		SequenceTree<?, ?> other = (SequenceTree<?, ?>) obj;
		if (finalNodes == null) {
			if (other.finalNodes != null)
				return false;
		} else if (!finalNodes.equals(other.finalNodes))
			return false;
		if (isLinked != other.isLinked)
			return false;
		if (sequenceValue == null) {
			if (other.sequenceValue != null)
				return false;
		} else if (!sequenceValue.equals(other.sequenceValue))
			return false;
		return true;
	}

	protected Map<S, SequenceTree<S, E>> createSubTreeMapInstance() {

		return new HashMap<S, SequenceTree<S, E>>();
	}

	/**
	 * Reconstitute the <tt>OrderedSequenceTree</tt> instance from a stream
	 * (i.e., deserialize it).
	 * 
	 * @see #writeObject
	 */
	@SuppressWarnings("unchecked")
	@Override
	public void readExternal(ObjectInput in) throws IOException,
			ClassNotFoundException {

		SequenceTree<S, E> readNode;
		Map<Integer, SequenceTree<S, E>> nodes = new HashMap<Integer, SequenceTree<S, E>>();
		Map<Integer, Set<Integer>> idSubTrees;
		Integer nodeCount;

		idSubTrees = (Map<Integer, Set<Integer>>) in.readObject();

		// This object properties
		readAndAssignToThis(in);

		// Node count, not including this node
		nodeCount = in.readInt();

		// Linked nodes
		for (int i = 0; i < nodeCount; i++) {
			readNode = readTreeSequentially(in);
			nodes.put(readNode.serialID, readNode);
		}

		// Restoring links
		SequenceTree<S, E> mapNode;
		this.subtrees = createSubTreeMapInstance();
		for (Integer id : idSubTrees.get(this.serialID)) {
			mapNode = nodes.get(id);
			mapNode.parent = this;
			this.subtrees.put(mapNode.getSequenceValue(), mapNode);
		}

		for (SequenceTree<S, E> ost : nodes.values()) {
			ost.subtrees = createSubTreeMapInstance();
			for (Integer id : idSubTrees.get(ost.serialID)) {
				mapNode = nodes.get(id);
				mapNode.parent = ost;
				ost.subtrees.put(mapNode.getSequenceValue(), mapNode);
			}
		}

	}

	/**
	 * Save the state of the <tt>OrderedSequenceTree</tt> instance to a stream
	 * (i.e., serialize it).
	 * 
	 * @serialData The <i>subtree map</i> of the {@code OrderedSequenceTree}
	 *             (the map whose keys are the nodes' serialID fields ({@code
	 *             Integer}) and the values the {@code Set<Integer>} containing
	 *             all the subtrees's serialID fields) is emitted ({@code
	 *             HashMap}), this class fields (serialID, isLinked,
	 *             sequenceValue and finalNodes) and next by the node count(
	 *             {@code Integer} all the {@code OrderedSequenceTree} objects
	 *             linked to this tree sequentially.
	 */
	@Override
	public void writeExternal(ObjectOutput out) throws IOException {
		Map<Integer, Set<Integer>> idSubtreesMap = new HashMap<Integer, Set<Integer>>();
		Integer nodeCount = 0;

		// Retrieving structure information and storing nodes
		Set<Integer> idSubTrees = new HashSet<Integer>();

		for (SequenceTree<S, E> ost : this.subtrees.values())
			idSubTrees.add(ost.serialID);

		idSubtreesMap.put(this.serialID, idSubTrees);

		for (SequenceTree<S, E> ost : this) {
			nodeCount++;

			idSubTrees = idSubtreesMap.get(ost.serialID);
			if (idSubTrees == null) {
				idSubTrees = new HashSet<Integer>();
				idSubtreesMap.put(ost.serialID, idSubTrees);
			}

			for (SequenceTree<S, E> sub : ost.subtrees.values()) {
				idSubTrees.add(sub.serialID);
			}
		}

		// Subtrees' map
		out.writeObject(idSubtreesMap);

		// first, this node properties
		writeTreeSequentially(out, this);

		// Node count, not including this node
		out.writeInt(nodeCount);

		// Linked nodes
		for (SequenceTree<S, E> ost : this) {
			writeTreeSequentially(out, ost);
		}
	}

	/**
	 * Parent field is not stored
	 */
	protected void writeTreeSequentially(ObjectOutput out,
			SequenceTree<S, E> ost) throws IOException {

		out.writeBoolean(ost.isLinked);
		out.writeInt(ost.serialID);
		out.writeObject(ost.sequenceValue);
		out.writeObject(ost.finalNodes);
	}

	/**
	 * Parent field is not retrieved
	 */
	@SuppressWarnings( { "unchecked" })
	protected SequenceTree<S, E> readTreeSequentially(ObjectInput in)
			throws IOException, ClassNotFoundException {

		SequenceTree<S, E> ost = new SequenceTree<S, E>(in
				.readBoolean());
		ost.serialID = in.readInt();
		ost.sequenceValue = (S) in.readObject();
		ost.finalNodes = (List<E>) in.readObject();

		return ost;
	}

	protected void readAndAssignToThis(ObjectInput in) throws IOException, ClassNotFoundException {

		SequenceTree<S, E> readNode = readTreeSequentially(in);
		this.parent = null;
		this.finalNodes = readNode.finalNodes;
		this.isLinked = readNode.isLinked;
		this.sequenceValue = readNode.sequenceValue;
		this.serialID = readNode.serialID;
	}

	private class TreeIterator implements Iterator<SequenceTree<S, E>> {

		private SequenceTree<S, E> currentNode;
		private Iterator<SequenceTree<S, E>> currentIterator;
		private Queue<SequenceTree<S, E>> toIterateQueue;

		public TreeIterator(SequenceTree<S, E> node) {

			currentNode = node;
			currentIterator = currentNode.subtrees.values().iterator();
			toIterateQueue = new LinkedList<SequenceTree<S, E>>();
		}

		@Override
		public boolean hasNext() {

			boolean hasNext = currentIterator.hasNext();

			while (!hasNext && currentNode != null) {
				currentNode = toIterateQueue.poll();
				if (currentNode != null) {
					currentIterator = currentNode.subtrees.values().iterator();
					hasNext = currentIterator.hasNext();
				}
			}

			return hasNext;
		}

		@Override
		public SequenceTree<S, E> next() {

			if (!hasNext())
				throw new NoSuchElementException("No more elements");

			SequenceTree<S, E> next = currentIterator.next();

			toIterateQueue.offer(next);

			return next;
		}

		@Override
		public void remove() {

			SequenceTree<S, E> next = next();

			next.getParent().subtrees.remove(next.sequenceValue);
		}
	}
}
