//    Copyright (C) 2010  Luis Alfonso Arce González, ajaest[@]gmail.com
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

package net.ajaest.lib.data;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.NavigableSet;
import java.util.TreeSet;

/**
 * Search tree by sequence. This class allows to sort a list of node objects
 * using a unique sequence of other objects. Every level of the tree represents
 * all the node objects that have in common a particular list in specific order
 * of other objects. The list of final nodes is an ordered sequence stored in a
 * {@code TreeSet}
 * 
 * @author Luis Alfonso Arce González
 * 
 * @param <S>
 *            the class of the sequence objects
 * @param <E>
 *            the class of the stored nodes objects
 */
public class SortedOrderedSequenceTree<S, E extends Comparable<E>> {

	// Structure
	private SortedOrderedSequenceTree<S, E> 		parent;
	private S 										sequenceValue;
	private Map<S, SortedOrderedSequenceTree<S, E>> subtrees;
	private NavigableSet<E> 						finalNodes;

	/**
	 * Creates a ordered sequence tree using the object default comparator for
	 * final nodes' set
	 * 
	 */
	public SortedOrderedSequenceTree() {



		subtrees = new HashMap<S, SortedOrderedSequenceTree<S, E>>();

		finalNodes = new TreeSet<E>();

		sequenceValue = null;
		parent = null;

	}

	/**
	 * Creates needed sequence tree if necessary in order to add a final node to
	 * the tree. If the {@code sequence.size()==0}, the final node is added to
	 * this tree.
	 * 
	 * @param sequence
	 *            the sorting sequence
	 * @param finalNode
	 *            the final node
	 */
	public void add(List<S> sequence, E finalNode) {

		if (sequence.size() > 0) {

			SortedOrderedSequenceTree<S, E> st = subtrees.get(sequence.get(0));

			if (st == null) {
				st = new SortedOrderedSequenceTree<S, E>();
				st.parent = this;
				st.sequenceValue = sequence.get(0);
				subtrees.put(sequence.get(0), st);
			}

			if (sequence.size() == 1) {
				st.addFinalNode(finalNode);
			} else {
				st.add(sequence.subList(1, sequence.size()), finalNode);
			}
		} else {
			this.addFinalNode(finalNode);
		}

	}

	/**
	 * Returns an unmodifiable map that maps subtrees from this tree by his
	 * sequence value.
	 * 
	 * @return a map with this tree's subtrees
	 */
	public Map<S, SortedOrderedSequenceTree<S, E>> getSubTrees() {

		return Collections.unmodifiableMap(subtrees);
	}

	/**
	 * 
	 * @return the list of final nodes related to this tree
	 */
	public NavigableSet<E> getFinalNodes() {

		return finalNodes;
	}

	/**
	 * Returns the subtree whose sequence value equals the specified one.
	 * 
	 * @param sequenceValue
	 *            the sequence value of the requested subtree
	 * @return the requested subtree, null otherwise
	 */
	public SortedOrderedSequenceTree<S, E> getSubTree(S sequenceValue) {

		return subtrees.get(sequenceValue);
	}

	/**
	 * 
	 * @return the sequence value that maps this tree
	 */
	public S getSequenceValue() {

		return sequenceValue;
	}

	/**
	 * 
	 * @return the tree parent of this tree
	 */
	public SortedOrderedSequenceTree<S, E> getParent() {

		return parent;
	}

	/**
	 * Adds a subtree to this tree using either provided {@code sequenceValue}
	 * or {@code subtree.sequenceValue}. If {@code sequenceValue} don't equals
	 * {@code null}, this value will be used to map the subtree. {@code
	 * subtree.sequenceValue} will be used to map the subtree if {@code
	 * sequenceValue} equals {@code null} only if {@code subtree.sequenceValue}
	 * does not equals {@code null}. If both sequence values equals {@code null}
	 * exception will be thrown. <br>
	 * Keep in mind that if subtree was nested in another tree before adding to
	 * this, previous parent information and sequence value will be overwritten
	 * in order to nest it to this structure.
	 * 
	 * @param sequenceValue
	 *            order mapping value, if null {@code sequenceValue} will be
	 *            used
	 * @param subtree
	 *            subtree to be added
	 * @throws NullPointerException
	 *             if {@code sequenceValue} and {@code subtree.sequenceValue}
	 *             are both null
	 */
	public void addSubTree(S sequenceValue, SortedOrderedSequenceTree<S, E> subtree) {

		S tempSV = sequenceValue != null ? sequenceValue : subtree.sequenceValue;

		if (tempSV == null)
			throw new NullPointerException("sortvalue and subtree.getSortValue() equals null, no sort information");

		if (subtrees.get(tempSV) != null)
			subtrees.get(tempSV).subtrees.get(tempSV).finalNodes.addAll(subtree.getFinalNodes());
		else {
			subtree.parent = this;
			subtree.sequenceValue = tempSV;
			subtrees.put(tempSV, subtree);
		}
	}

	/**
	 * Appends the specified element to the end of this list.
	 * 
	 * @param node
	 *            element to be appended to this list
	 * @return <tt>true</tt> (as specified by {@link Collection#add})
	 */
	public boolean addFinalNode(E node) {

		return finalNodes.add(node);
	}

	/**
	 * Removes the mapping for the specified key from this map if present.
	 * 
	 * @param sequenceValue
	 *            key whose mapping is to be removed from the map
	 * @return the previous value associated with <tt>key</tt>, or <tt>null</tt>
	 *         if there was no mapping for <tt>key</tt>.
	 */
	public SortedOrderedSequenceTree<S, E> removeSubTree(S sequenceValue) {

		return subtrees.remove(sequenceValue);
	}

	/**
	 * Removes the first occurrence of the specified element from this list, if
	 * it is present. If the list does not contain the element, it is unchanged.
	 * More formally, removes the element with the lowest index <tt>i</tt> such
	 * that
	 * <tt>(o==null&nbsp;?&nbsp;get(i)==null&nbsp;:&nbsp;o.equals(get(i)))</tt>
	 * (if such an element exists). Returns <tt>true</tt> if this list contained
	 * the specified element (or equivalently, if this list changed as a result
	 * of the call).
	 * 
	 * @param node
	 *            element to be removed from this list, if present
	 * @return <tt>true</tt> if this list contained the specified element
	 */
	public boolean removeFinalNode(E node) {

		return finalNodes.remove(node);
	}

	/**
	 * Does a recursive search along subtrees of this tree in order to get final
	 * nodes that matches the specified sequence. If not matching final node is
	 * found, it returns an empty list.
	 * 
	 * @param sequence
	 *            sequence to be searched
	 * @return list of final nodes that matches the sequence
	 */
	public NavigableSet<E> search(List<S> sequence) {

		if (sequence.size() == 0)
			return finalNodes;
		else {
			SortedOrderedSequenceTree<S, E> temp = subtrees.get(sequence.get(0));
			if (temp != null) {
				if (sequence.size() == 1)
					return temp.finalNodes;
				else
					return temp.search(sequence.subList(1, sequence.size()));
			}
		}

		return new TreeSet<E>();

	}

	/**
	 * Says whether exists or not any final node in this tree referenced by the
	 * specified sequence
	 * 
	 * @param sequence
	 *            sequence to be checked
	 * @return true if {@code search(sequence)} does not returns an empty list,
	 *         false otherwise
	 */
	public boolean exists(List<S> sequence) {

		return search(sequence).size() != 0;

	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((finalNodes == null) ? 0 : finalNodes.hashCode());
		result = prime * result
				+ ((sequenceValue == null) ? 0 : sequenceValue.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (!(obj instanceof SortedOrderedSequenceTree<?, ?>)) {
			return false;
		}
		SortedOrderedSequenceTree<?, ?> other = (SortedOrderedSequenceTree<?, ?>) obj;
		if (finalNodes == null) {
			if (other.finalNodes != null) {
				return false;
			}
		} else if (!finalNodes.equals(other.finalNodes)) {
			return false;
		}
		if (sequenceValue == null) {
			if (other.sequenceValue != null) {
				return false;
			}
		} else if (!sequenceValue.equals(other.sequenceValue)) {
			return false;
		}
		return true;
	}

	
}
