package net.ajaest.lib.swing.util;

import java.awt.Component;
import java.awt.Font;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.HashMap;
import java.util.Map;

import javax.swing.Action;
import javax.swing.ActionMap;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.text.DefaultEditorKit;
import javax.swing.text.JTextComponent;

import net.ajaest.jdk.core.winHandlers.WinHandler;
import net.ajaest.lib.locale.SwingMessages;
import net.ajaest.lib.locale.SwingMessagesEnums;
import net.ajaest.lib.locale.SwingMessagesEnums.ContextMenuEnums.editMenu;
import net.ajaest.lib.swing.util.action.JComboBoxCopyTextAction;
import net.ajaest.lib.swing.util.action.JLabelCopyTextAction;
import net.ajaest.lib.utils.SystemEnums.ISO639ー1;

public class EditContextMenuEngine implements WinHandler {

	private static final long serialVersionUID = -5221367951857450395L;

	private Map<Component, JPopupMenu> menus;
	private Map<SwingMessagesEnums.ContextMenuEnums.editMenu, String> messages;
	private ContextMenuMouseListener cmml;

	public EditContextMenuEngine(ISO639ー1 messageLang) {
		messages = SwingMessages.getLocalizedMessages(messageLang).getContextMenuEditMessages();
		menus = new HashMap<Component, JPopupMenu>();
		cmml = new ContextMenuMouseListener(this);
	}

	/**
	 * Adds manually an edit context menu to the specified component.
	 * 
	 * @param c
	 *            the component to which context menu will be added.
	 * @param jpm
	 *            the pop menu to be added
	 * @return the added {@code JPopupMenu}, null if no context menu were added.
	 */
	public JPopupMenu addContextMenu(JComponent c, JPopupMenu jpm) {

		if (jpm != null) {
			c.add(jpm);
			menus.put(c, jpm);
		}

		return jpm;
	}

	/**
	 * Adds a default edit context menu to the specified component.
	 * 
	 * @param c
	 *            the component to which context menu will be added.
	 * @return the added {@code JPopupMenu}, null if no context menu were added.
	 */
	public JPopupMenu addContextMenu(JComponent c) {
		JPopupMenu jpm = createDefaultEditContextMenu(c);
		if (jpm != null) {
			c.add(jpm);
			menus.put(c, jpm);
		}

		return jpm;
	}

	/**
	 * Returns {@code null} if no default edit context menu can be created
	 */
	private JPopupMenu createDefaultEditContextMenu(Component c) {

		JPopupMenu jpm = null;

		if (c instanceof JTextComponent) {

			jpm = new JPopupMenu();

			JTextComponent jtc = (JTextComponent) c;

			ActionMap actionMap = jtc.getActionMap();
			Action copyAction = actionMap.get(DefaultEditorKit.copyAction);
			Action pasteAction = actionMap.get(DefaultEditorKit.pasteAction);
			Action cutAction = actionMap.get(DefaultEditorKit.cutAction);
			Action selectAllAction = actionMap.get(DefaultEditorKit.selectAllAction);

			JMenuItem jmCut = new JMenuItem(cutAction);
			jmCut.setText(messages.get(editMenu.CUT));
			jpm.add(jmCut);

			JMenuItem jmCopy = new JMenuItem(copyAction);
			jmCopy.setText(messages.get(editMenu.COPY));
			jpm.add(jmCopy);

			JMenuItem jmPaste = new JMenuItem(pasteAction);
			jmPaste.setText(messages.get(editMenu.PASTE));
			jpm.add(jmPaste);

			JMenuItem jmSelectAll = new JMenuItem(selectAllAction);
			jmSelectAll.setText(messages.get(editMenu.SELECT_ALL));
			jpm.add(jmSelectAll);

			jtc.addMouseListener(cmml);
		} else if(c instanceof JLabel) {
			
			jpm = new JPopupMenu();
			JLabel jl = (JLabel)c;
			
			JMenuItem jmCopy = new JMenuItem(new JLabelCopyTextAction(jl));
			jmCopy.setText(messages.get(editMenu.COPY));
			jpm.add(jmCopy);

			jl.addMouseListener(cmml);
		} else if (c instanceof JComboBox) {

			jpm = new JPopupMenu();

			JComboBox jcb = (JComboBox) c;

			Action copyAction = new JComboBoxCopyTextAction(jcb);
			
			JMenuItem jmCopy = new JMenuItem(copyAction);
			jmCopy.setText(messages.get(editMenu.COPY));
			jpm.add(jmCopy);

			JTextComponent editableCBComp = (JTextComponent) jcb.getEditor().getEditorComponent();
			
			editableCBComp.addMouseListener(cmml);
			jcb.addMouseListener(cmml);
		}

		return jpm;
	}

	public JPopupMenu getContextMenu(Component c) {
		return menus.get(c);
	}

	private class ContextMenuMouseListener extends MouseAdapter {

		private EditContextMenuEngine ecm;

		private ContextMenuMouseListener(EditContextMenuEngine ecm) {
			this.ecm = ecm;
		}

		@Override
		public void mouseReleased(MouseEvent e) {
			showContextMenu(e);
		}

		@Override
		public void mousePressed(MouseEvent e) {
			showContextMenu(e);
		}

		private void showContextMenu(MouseEvent e) {

			JPopupMenu jpm = ecm.getContextMenu(e.getComponent());

			if (jpm == null) {
				jpm = ecm.createDefaultEditContextMenu(e.getComponent());
				ecm.menus.put(e.getComponent(), jpm);
			}

			if (e.isPopupTrigger() && jpm != null) {
				e.getComponent().requestFocus();
				jpm.show(e.getComponent(), e.getX(), e.getY());
				jpm.setVisible(true);
			}
		}
	}

	
	// TODO: add to API
	@Override
	public Font getUnicodeFont() {
		
		return null;
	}
}
