//    Copyright (C) 2010  Luis Alfonso Arce González, ajaest[@]gmail.com
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

package net.ajaest.lib.swing.util;

import static java.lang.Math.PI;

import java.awt.geom.Line2D;
import java.util.ArrayList;
import java.util.List;

import net.ajaest.jdk.data.auxi.AllowedStrokeLineEnum;
import net.ajaest.lib.data.Pair;

//TODO: JavaDoc
public class JLineDraw {

	/**
	 * {@code ·····\++++|++++/·····}<br>
	 * {@code ······\+++|+++/······}<br>
	 * {@code ·······\++|++/·······}<br>
	 * {@code ........\+|+/........}<br>   
	 * {@code _________\|/_________}<br>
	 *       
	 * Represents the maximum horizontal line angle tolerance value.
	 * 
	 */
	public static double horAngleTolerance = PI / 20;
	/**
	 * {@code +++\····|····/+++}<br>
	 * {@code ++++\···|···/++++}<br>
	 * {@code +++++\··|··/+++++}<br>
	 * {@code ++++++\·|·/++++++}<br>   
	 * {@code +++++++\|/+++++++}<br>
	 * {@code --------'--------}<br>      
	 *  
	 * Represents the maximum horizontal line angle value; 
	 * 
	 */
	public static double verAngleTolerance = PI / 20;
	
	private List<Line2D> lines;

	private List<Pair<Integer, Integer>> strokes;

	public JLineDraw(List<Integer> x1, List<Integer> y1, List<Integer> x2, List<Integer> y2, List<Pair<Integer, Integer>> strokes) {
		
		if (!checkIntegrity(x1, y1, x2, y2, strokes)) {
			throw new IllegalArgumentException("There doesn't exist integrity between lines and strokes definitions");
		}
		
		lines = new ArrayList<Line2D>();
		this.strokes = strokes;
		
		for (int i = 0; i < x1.size(); i++) {
			// minus in order to convert from graphics context's coordinate
			// system into Cartesian
			lines.add(new Line2D.Float(x1.get(i), -y1.get(i), x2.get(i), -y2.get(i)));
		}

	}
	
	public static AllowedStrokeLineEnum lineToAsteriskModel(Line2D line) {

		Double slope = ((line.getY2() - line.getY1())) / ((line.getX2() - line.getX1()));
		Double angle = Math.atan(slope);
		if (angle < 0)
			angle += PI;

		Boolean leftToRight = (line.getX1() - line.getX2()) < 0;
		Boolean upToDown = (line.getY1() - line.getY2()) > 0;
		
		AllowedStrokeLineEnum lineType;

		//horizontal
		if (0 <= angle && angle <= (0 + horAngleTolerance) || (PI - horAngleTolerance) <= angle && angle <= PI) {
			if (leftToRight)
				lineType = AllowedStrokeLineEnum.F;
			else
				lineType = AllowedStrokeLineEnum.D;
		}
		// positive slope
		else if ((0 + horAngleTolerance) < angle && angle < (PI / 2 - verAngleTolerance)) {
			if (leftToRight)
				lineType = AllowedStrokeLineEnum.I;
			else
				lineType = AllowedStrokeLineEnum.A;
		}
		// vertical
		else if ((PI / 2 - verAngleTolerance) <= angle && angle <= (PI / 2 + verAngleTolerance)) {
			if (upToDown)
				lineType = AllowedStrokeLineEnum.B;
			else
				lineType = AllowedStrokeLineEnum.H;
		}
		// negative slope
		else if ((PI / 2 + verAngleTolerance) < angle && angle < (PI - verAngleTolerance)) {
			if (leftToRight)
				lineType = AllowedStrokeLineEnum.C;
			else
				lineType = AllowedStrokeLineEnum.G;
		} else
			lineType = AllowedStrokeLineEnum.E;

		return lineType;
	}

	public List<AllowedStrokeLineEnum> getAsteriskModel() {

		List<AllowedStrokeLineEnum> strokes = new ArrayList<AllowedStrokeLineEnum>();

		AllowedStrokeLineEnum astLine;
		for (Line2D line : lines) {
			astLine = lineToAsteriskModel(line);
			strokes.add(astLine);
		}

		return strokes;
	}

	private boolean checkIntegrity(List<Integer> x1, List<Integer> y1, List<Integer> x2, List<Integer> y2, List<Pair<Integer, Integer>> strokes) {

		boolean integrity = x1.size() == y1.size() && x1.size() == x2.size() && x1.size() == y2.size();

		if (integrity) {

			Integer size = x1.size();

			integrity = integrity && strokes.get(0).getFirst() < size;
			integrity = integrity && strokes.get(0).getSecond() < size;
			Integer prevBeginStroke = strokes.get(0).getFirst();
			Integer prevEndStroke = strokes.get(0).getSecond();

			for (int i = 1; i < strokes.size() && integrity; i++) {

				integrity = integrity && strokes.get(i).getFirst() < size;
				integrity = integrity && strokes.get(i).getSecond() <= size;
				integrity = integrity && strokes.get(i).getFirst() > prevBeginStroke;
				integrity = integrity && strokes.get(i).getSecond() > prevEndStroke;
				prevBeginStroke = strokes.get(i).getFirst();
				prevEndStroke = strokes.get(i).getSecond();
			}
		}

		return integrity;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((lines == null) ? 0 : lines.hashCode());
		result = prime * result + ((strokes == null) ? 0 : strokes.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		JLineDraw other = (JLineDraw) obj;
		if (lines == null) {
			if (other.lines != null)
				return false;
		} else if (!lines.equals(other.lines))
			return false;
		if (strokes == null) {
			if (other.strokes != null)
				return false;
		} else if (!strokes.equals(other.strokes))
			return false;
		return true;
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append("JLineDraw [lines=");
		builder.append(lines);
		builder.append(", strokes=");
		builder.append(strokes);
		builder.append("]");
		return builder.toString();
	}


	public static void setHorizontalTolerance(Double tolerance) {
		horAngleTolerance = Math.abs(tolerance);
	}

	public static void setVerticalTolerance(Double tolerance) {
		verAngleTolerance = Math.abs(tolerance);
	}

	public static Double getHorizontalTolerance() {
		return horAngleTolerance;
	}

	public static Double getVerticalTolerance() {
		return verAngleTolerance;
	}

	public List<Pair<Integer, Integer>> getStrokes() {
		return strokes;
	}

}
